// shadowproj.cpp
// by Glenn G. Chappell
// November 2003
//
// For CS 481/681
// A simple OpenGL/GLUT program that demonstrates shadow
// generation using projection.
// Displays a rotating cube and its shadow on a plane.
//
// Potions of this program based on dinoshade.c
// Copyright (c) Mark J. Kilgard, 1994, 1997.
// Available on MJK's "Way cool way fast OpenGL rendering techniques" web page
// http://www.opengl.org/developers/code/mjktips/

#include <iostream>
using std::cerr;
using std::endl;
#include <string>
using std::string;
#include <math.h>  // Some versions of MS-Vis C++ have broken <cmath>
//using std::sqrt;
#include <stdlib.h>  // Some versions of MS-Vis C++ have broken <cstdlib>
//using std::exit;
#include <GL/glut.h> // GLUT stuff - includes OpenGL headers as well


// Global variables
// Window/viewport
const int startwinsize = 400; // Starting window width & height, in pixels
const double neardist = 1.0;  // Near & far clipping distances
const double fardist = 10.0;

// Keyboard
const int ESCKEY = 27;        // ASCII value of escape character

// Scene
bool drawhelp = true;         // True if instructions shown

double angle = 0.0;           // Current rotation angle for object
const double anglestep = 1.0;
bool rotate = false;          // True if rotating
bool drawshadow = true;       // True if shadow drawn
bool drawallfaces = true;     // True if all faces drawn; otherwise opening in object

GLdouble lightpos[] = { 0.0, 4.0, 0.0, 1.0 };  // Light position in homogeneous form
const double plane_size = 3.0;                 // Extent of plane
const double plane_level = -2.5;               // Level (y-coord) of plane
GLdouble planepts[4][3] = {                    // Corners of "plane"
   { -plane_size, plane_level, -plane_size },
   { -plane_size, plane_level,  plane_size },
   {  plane_size, plane_level,  plane_size },
   {  plane_size, plane_level, -plane_size } };

GLdouble plane_eq[4];                          // Holds plane "equation"
                                               //  in the form Ax+By+Cz=D
GLdouble shadowmat[16];                        // Shadow projection matrix


// printbitmap
// Prints the given string at the given raster position
//  using GLUT bitmap fonts.
// You probably don't want any rotations in the model/view
//  transformation when calling this function.
void printbitmap(const string msg, double x, double y)
{
   glRasterPos2d(x, y);
   for (string::const_iterator ii = msg.begin();
        ii != msg.end();
        ++ii)
   {
      glutBitmapCharacter(GLUT_BITMAP_9_BY_15, *ii);
   }
}


// makeshadowmatrix
// Create a matrix that will project the desired shadow.
// Is given the plane equation, in the form Ax+By+Cz=D.
// Is given light postion, in homogeneous form.
// Creates matrix in shadowmat.
void makeshadowmatrix(GLdouble shadowmat[16],
                      const GLdouble plane_eq[4],
                      const GLdouble lightpos[4])
{
// Find dot product between light position vector and ground plane normal.
   GLdouble dot = plane_eq[0] * lightpos[0] +
                  plane_eq[1] * lightpos[1] +
                  plane_eq[2] * lightpos[2] +
                  plane_eq[3] * lightpos[3];

   typedef GLdouble (*mat44)[4];
   mat44 sm = reinterpret_cast<mat44>(shadowmat);

   sm[0][0] = dot - lightpos[0] * plane_eq[0];
   sm[1][0] =     - lightpos[0] * plane_eq[1];
   sm[2][0] =     - lightpos[0] * plane_eq[2];
   sm[3][0] =     - lightpos[0] * plane_eq[3];

   sm[0][1] =     - lightpos[1] * plane_eq[0];
   sm[1][1] = dot - lightpos[1] * plane_eq[1];
   sm[2][1] =     - lightpos[1] * plane_eq[2];
   sm[3][1] =     - lightpos[1] * plane_eq[3];
   
   sm[0][2] =     - lightpos[2] * plane_eq[0];
   sm[1][2] =     - lightpos[2] * plane_eq[1];
   sm[2][2] = dot - lightpos[2] * plane_eq[2];
   sm[3][2] =     - lightpos[2] * plane_eq[3];

   sm[0][3] =     - lightpos[3] * plane_eq[0];
   sm[1][3] =     - lightpos[3] * plane_eq[1];
   sm[2][3] =     - lightpos[3] * plane_eq[2];
   sm[3][3] = dot - lightpos[3] * plane_eq[3];
}


// findplane
// Calculate A,B,C,D version of plane equation,
// given 3 non-colinear points in plane
void findplane(GLdouble plane_eq[4],
               const GLdouble p0[3],
               const GLdouble p1[3],
               const GLdouble p2[3])
{
    GLdouble vec0[3], vec1[3];

   // Need 2 vectors to find cross product.
   vec0[0] = p1[0] - p0[0];
   vec0[1] = p1[1] - p0[1];
   vec0[2] = p1[2] - p0[0];

   vec1[0] = p2[0] - p0[0];
   vec1[1] = p2[1] - p0[1];
   vec1[2] = p2[2] - p0[2];

   // find cross product to get A, B, and C of plane equation
   plane_eq[0] = vec0[1] * vec1[2] - vec0[2] * vec1[1];
   plane_eq[1] = vec0[2] * vec1[0] - vec0[0] * vec1[2];
   plane_eq[2] = vec0[0] * vec1[1] - vec0[1] * vec1[0];

   GLdouble normlen = sqrt(plane_eq[0]*plane_eq[0]
                         + plane_eq[1]*plane_eq[1]
                         + plane_eq[2]*plane_eq[2]);
   if (normlen != 0)
   {
      plane_eq[0] /= normlen;
      plane_eq[1] /= normlen;
      plane_eq[2] /= normlen;
   }
   else
   {
      plane_eq[0] = 1.;
      plane_eq[1] = 0.;
      plane_eq[2] = 0.;
   }

   plane_eq[3] = -(plane_eq[0] * p0[0]
                 + plane_eq[1] * p0[1]
                 + plane_eq[2] * p0[2]);
}


void drawplane()
{
   glColor3d(0.2, 0.2, 0.8);
   glBegin(GL_POLYGON);
      glVertex3dv(planepts[0]);
      glVertex3dv(planepts[1]);
      glVertex3dv(planepts[2]);
      glVertex3dv(planepts[3]);
   glEnd();
}


// Draws object
// In various colors if usecolors is set.
// Otherwise in gray.
// Colors are done with glColor*,
// so they will be ignored if lighting is enabled.
void drawobject(bool usecolors)
{
   const double h = 0.7;

   if (!usecolors)
      glColor3d(0.2, 0.2, 0.2);

   glBegin(GL_QUADS);
      if (usecolors)
         glColor3d(0.9, 0.1, 0.1);
      glVertex3d( h, h, h);
      glVertex3d(-h, h, h);
      glVertex3d(-h,-h, h);
      glVertex3d( h,-h, h);

      if (usecolors)
         glColor3d(0.1, 0.9, 0.9);
      glVertex3d( h, h,-h);
      glVertex3d( h,-h,-h);
      glVertex3d(-h,-h,-h);
      glVertex3d(-h, h,-h);

      if (usecolors)
         glColor3d(0.1, 0.9, 0.1);
      glVertex3d( h, h, h);
      glVertex3d( h,-h, h);
      glVertex3d( h,-h,-h);
      glVertex3d( h, h,-h);

      if (usecolors)
         glColor3d(0.9, 0.1, 0.9);
      glVertex3d(-h, h, h);
      glVertex3d(-h, h,-h);
      glVertex3d(-h,-h,-h);
      glVertex3d(-h,-h, h);

      if (drawallfaces)
      {
         if (usecolors)
            glColor3d(0.1, 0.1, 0.9);
         glVertex3d( h, h, h);
         glVertex3d( h, h,-h);
         glVertex3d(-h, h,-h);
         glVertex3d(-h, h, h);

         if (usecolors)
            glColor3d(0.9, 0.9, 0.1);
         glVertex3d( h,-h, h);
         glVertex3d(-h,-h, h);
         glVertex3d(-h,-h,-h);
         glVertex3d( h,-h,-h);
      }
   glEnd();
}


// display
// The GLUT display function
void display()
{
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);  // Clear the viewport

   // Draw the plane
   glPushMatrix();
      glTranslated(0., 0., -4.);
      drawplane();
   glPopMatrix();

   // Draw the shadow, if enabled
   if (drawshadow)
   {
      glDisable(GL_DEPTH_TEST);
      glPushMatrix();
         glTranslated(0., 0., -4.);
         glMultMatrixd(shadowmat);
         glRotated(angle, 0.5, 1.0, 0.0);
         drawobject(false);  // gray
      glPopMatrix();
      glEnable(GL_DEPTH_TEST);
   }

   // Draw the object
   glPushMatrix();
      glTranslated(0., 0., -4.);
      glRotated(angle, 0.5, 1.0, 0.0);
      drawobject(true);  // colored
   glPopMatrix();

   // Draw instructions, if enabled
   if (drawhelp)
   {
      glDisable(GL_DEPTH_TEST);
      glMatrixMode(GL_PROJECTION);
      glPushMatrix();  // Save old projection matrix
         glLoadIdentity();
         gluOrtho2D(-1., 1., -1., 1.);  // old faithful
         glColor3d(1.0, 1.0, 1.0);
         printbitmap("Space Toggle rotation", -0.9, 0.9);
         printbitmap("S     Toggle shadow", -0.9, 0.8);
         printbitmap("O     Toggle opening in cube", -0.9, 0.7);
         printbitmap("H     Toggle help", -0.9, 0.6);
         printbitmap("Esc   Quit", -0.9, 0.5);
      glPopMatrix();
      glMatrixMode(GL_MODELVIEW);  // Always go back to model/view mode
      glEnable(GL_DEPTH_TEST);
   }

   glutSwapBuffers();
}


// idle
// The GLUT idle function
void idle()
{
   // Print OpenGL errors, if there are any (for debugging)
   if (GLenum err = glGetError())
   {
      cerr << "OpenGL ERROR: " << gluErrorString(err) << endl;
   }

   // Do rotation
   if (rotate)
   {
      angle += anglestep;
      glutPostRedisplay();
   }
}


// reshape
// The GLUT reshape function
void reshape(int w, int h)
{
   glViewport(0, 0, w, h);

   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   gluPerspective(90, double(w)/h, neardist, fardist);

   glMatrixMode(GL_MODELVIEW);  // Always go back to modelview mode
}


// keyboard
// The GLUT keyboard function
void keyboard(unsigned char key, int x, int y)
{
   switch (key)
   {
   case ESCKEY:  // ESC: Quit
      exit(0);
      break;
   case ' ':     // Spc: Toggle rotation
      rotate = !rotate;
      break;
   case 's':     // S: Toggle shadow
   case 'S':
      drawshadow = !drawshadow;
      glutPostRedisplay();
      break;
   case 'o':     // O: Toggle opening
   case 'O':
      drawallfaces = !drawallfaces;
      glutPostRedisplay();
      break;
   case 'h':     // H: toggle help
   case 'H':
      drawhelp = !drawhelp;
      glutPostRedisplay();
      break;
   }
}


// init
// Initialization
// Called by main
void init()
{
   glEnable(GL_DEPTH_TEST);  // We're doing 3-D
   glClearColor(0., 0., 0., 0.);  // Background color

   // Make plane equation vector & shadow matrix
   findplane(plane_eq, planepts[0], planepts[1], planepts[2]);
   makeshadowmatrix(shadowmat, plane_eq, lightpos);
}


int main(int argc, char ** argv)
{
   // Initialize OpenGL/GLUT
   glutInit(&argc, argv);
   glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);

   // Make a window
   glutInitWindowSize(startwinsize, startwinsize);
   glutInitWindowPosition(50, 50);
   glutCreateWindow("CS 481/681 - Shadowing via Projection");

   // Initialize GL states & register callbacks
   init();
   glutDisplayFunc(display);
   glutReshapeFunc(reshape);
   glutKeyboardFunc(keyboard);
   glutIdleFunc(idle);

   // Do something
   glutMainLoop();

   return 0;
}

